<?php
/**
*    This file is part of EvalWF Module - Automatic evaluation of forms was made by webform module.
*    Copyright (C) 2020-2023  University of Szeged
*
*    This EvalWF Module is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    Foobar is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Foobar.  If not, see <https://www.gnu.org/licenses/>.
*    it under the terms of the GNU General Public License as published by
*
* @file \Drupal\evalwf\Controller\EvalWFController.php - Creates user side content / the evaluation page
*
* @author Tóthpál István
*/

namespace Drupal\evalwf\Controller;

use Drupal\Component\Utility\Html;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Url;
use Drupal\Core\Mail\Plugin\Mail\PhpMail;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\evalwf\Entity\EvalWF;
use Drupal\evalwf\Entity\EvalWFEvaluation;
use Drupal\webform\Entity\Webform;
use Drupal\webform\Entity\WebformSubmission;
use Symfony\Component\HttpFoundation\RedirectResponse;

class EvalWFController {

  private $entity = null;


  /**
   *   Loads the selected Entity
   *
   *     - @params comes from route: evalwf/[webform:id]/[webform_submission:sid]
   *     or from urlquery: evalwf?wfid=[webform:id]&sid=[webform_submission:sid]
   *
   *   @param Webform weform                         - the selected webform
   *   @param WebformSubmission webform_submission   - the selected webform_submission
   *
   *   this->entity    - the loaded entity
   *   this->wfid      - the selected webform_id
   *   this->sid       - the selected webform_submission_id
   *   this->webform_elements    - the form elements in the selected webform
   *   this->webform_submission  - the selected webform_submission
   */
  function LoadEntity( Webform $webform = null, WebformSubmission $webform_submission = null) {
    if ($webform && $webform_submission) {
      if ($webform->id() == $webform_submission->bundle()) {
        $this->wfid = $webform->id();
      }
      else {
        $this->wfid = $webform_submission->bundle();
        $webform = Webform::load($this->wfid);
        \Drupal::messenger()->addMessage( t('Different webform connected to selected submission, evaluated with submission\'s webform.'), 'warning');
      }
      $this->sid = $webform_submission->id();
    }
    else {
      $urlquery = \Drupal::request()->query->all();
      if (isset($urlquery['sid'])) $this->sid=$urlquery['sid'];
      if (isset($urlquery['wfid'])) $this->wfid=$urlquery['wfid'];
    }

    if (isset($this->sid) && isset($this->wfid)) {
      if(!$webform) $webform = Webform::load($this->wfid);
      $this->webform_elements = $webform->getElementsDecoded();
      $this->webform_submission = WebformSubmission::load($this->sid);
    }
    if (isset($this->wfid)) {
      $this->entity = EvalWF::LoadEntityByWebformID($this->wfid);
    }
  }

  /**
   * Returns the evaluation page title. / HTML <title> - Drupal will append sitename/
   */
  public function getTitle( Webform $webform = null, WebformSubmission $webform_submission = null) {
    $this->LoadEntity($webform, $webform_submission);
    if ($this->entity) {
      return  $this->entity->getEvaluationTitle();
    }
  }

  /**
   *    Creates the elements tree for theme
   *
   *   @param array  elements   - webform elements
   *   @param array  data       - webform submission data
   *   @param array  settings   - the evaluation settings data
   *   @returns array
   *        array  childs           -  elements theme data
   *        string subtotal         -  total points of all elements
   *        array  elements_points  -  elements and its evaluated points
   */
  private function buildelementstree( $elements, $data, $settings ) {
    $tree = [];
    $points = 0;
    $elements_points = [];
    foreach ($elements as $key => $element ) {
      $subpoints = 0;
      if (isset($element['#type'])) {
        $tree[$key]['type'] = $element['#type'];
        $tree[$key]['title'] = ( isset($element['#title']) ? $element['#title'] : ( isset($element['#caption']) ? $element['#caption'] : '' ));
        $tree[$key]['description'] = ( isset($element['#description']) ? $element['#description'] : '' );
        $tree[$key]['header'] = [ ['data'=>'', 'attributes'=>' width=5'],t('answer'),t('points'), ];
        switch ($element['#type']) {
          case 'details':
          case 'fieldset':
          case 'container':
            $childs_data = $this->buildelementstree($element, $data, (isset($settings[$key]) ? $settings[$key] : null ));
            $tree[$key]['children'] = $childs_data['childs'];
            $subpoints += $childs_data['subtotal'];
            $tree[$key]['subtotal'] = t('Sum of @element: %subtotal Point(s)', [ '@element'=>$tree[$key]['title'], '%subtotal'=>$childs_data['subtotal']] );
            $elements_points = array_merge( $elements_points, array( $key => $childs_data['subtotal']) );
            $elements_points = array_merge( $elements_points, $childs_data['elements_points'] );
            break;
          case 'textarea':
          case 'textfield':
          case 'email':
          case 'webform_autocomplete':
            $tree[$key]['header'] = null;
            $tree[$key]['text'] = ( isset($data[$key]) ? $data[$key] : null );
            break;
          case 'checkbox':
            $tree[$key]['rows'] = [
              $key => [
                'selected' => ($data[$key]),
                'answer' => $element['#title'],
                'points'=> $settings[$key.'_t'][$key]['points'],
                'isgood'=> isset($data[$key]) && ($data[$key]),
              ],
             ];
            if (isset($data[$key]) && ($data[$key])) {
              $subpoints += $settings[$key.'_t'][$key]['points'];
            }
            $tree[$key]['footer'] = [[
              'data'=>t('@element: %subtotal Point(s)', [ '@element'=>$tree[$key]['title'], '%subtotal'=> $subpoints]),
              'attributes'=>' colspan=3',
            ]];
            $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
            break;
        case 'webform_checkboxes_other':
          $otherval = (isset($data[$key]) ? $data[$key] : [] );
          $ov = null;
          foreach($element['#options'] as $optkey => $optvalue ) {
            foreach( $otherval as $okey => $oval ) {
              if ($oval == $optkey) $otherval[$okey] = null;
            }
          }
          foreach( $otherval as $okey => $oval ) {
            if ($oval!=null )$ov = $oval;
          }
          $other = $this->buildelementstree(
            [ $key.'_other' => array(
              '#type' => ( isset($element['#other__type']) ? $element['#other__type'] : 'textfield' ),
              '#title' => ( isset($element['#other__title']) ? $element['#other__title'] : null ),
              '#description' => ( isset($element['##other__description']) ? $element['##other__description'] : null ),
              '#size' => ( isset($element['#other__size']) ? $element['#other__size'] : null ),
            )],
            [$key.'_other' => $ov],
            ( isset($settings[$key][$key.'_other']) ? [$key.'_other_t' => [ $key.'_other' => $settings[$key][$key.'_other']]] : [] )
          );
        case 'checkboxes':
        case 'tableselect':
          foreach($element['#options'] as $optkey => $optvalue ) {
            if (((!$this->entity->isShowAll()) && isset($data[$key]) && ( in_array($optkey,$data[$key]) )) || ($this->entity->isShowAll())) {
              $tree[$key]['rows'][$optkey] = [
                'selected' => (isset($data[$key]) ? in_array($optkey,$data[$key]) : FALSE ),
                'answer' => $optvalue,
                'points'=> (isset($settings[$key][$optkey]['points']) ? $settings[$key][$optkey]['points'] : 0 ),
                'isgood'=> isset($data[$key]) && in_array($optkey,$data[$key]),
              ];
            }
            if ( isset($data[$key]) && in_array($optkey,$data[$key]) ) {
              $subpoints += $settings[$key][$optkey]['points'];
            }
          }
          if (
            (isset($other['childs'][$key.'_other']['rows'][$key.'_other'])) &&
            (((!$this->entity->isShowAll()) && isset($data[$key]) && ( $ov!=null )) || ($this->entity->isShowAll()))
            ) {
              $tree[$key]['rows']['other'] = $other['childs'][$key.'_other']['rows'][$key.'_other'];
              $tree[$key]['rows']['other']['selected'] = ($ov!=null);
              $subpoints += $other['subtotal'];
            }
          $tree[$key]['footer'] = [[
            'data'=>t('@element: %subtotal Point(s)', [ '@element'=>$tree[$key]['title'], '%subtotal'=> $subpoints]),
            'attributes'=>' colspan=3'
          ]];
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
        case 'webform_radios_other':
        case 'webform_select_other':
          $other = $this->buildelementstree(
            [ $key.'_other' => array(
              '#type' => ( isset($element['#other__type']) ? $element['#other__type'] : 'textfield' ),
              '#title' => ( isset($element['#other__title']) ? $element['#other__title'] : null ),
              '#description' => ( isset($element['##other__description']) ? $element['##other__description'] : null ),
              '#size' => ( isset($element['#other__size']) ? $element['#other__size'] : null ),
            )],
            [$key.'_other' => (isset($data[$key]) && (!in_array($data[$key],$element['#options'])) ? $data[$key] : null )],
            ( isset($settings[$key][$key.'_other']) ? [$key.'_other_t' => [ $key.'_other' => $settings[$key][$key.'_other']]] : [] )
          );
        case 'radios':
        case 'select':
          foreach($element['#options'] as $optkey => $optvalue ) {
            if (((!$this->entity->isShowAll()) && isset($data[$key]) && ( $optkey==$data[$key] )) || ($this->entity->isShowAll())) {
              $tree[$key]['rows'][$optkey] = [
                'selected' => ((isset($data[$key]) && ($optkey==$data[$key])) ? TRUE : FALSE ),
                'answer' => $optvalue,
                'points'=> (isset($settings[$key][$optkey]['points']) ? $settings[$key][$optkey]['points'] : 0 ),
                'isgood'=> isset($data[$key]) && ($optkey==$data[$key]),
              ];
            }
            if ( isset($data[$key]) && ($optkey==$data[$key]) ) {
              $subpoints += $settings[$key][$optkey]['points'];
            };
          }
          if (
            (isset($other['childs'][$key.'_other']['rows'][$key.'_other'])) &&
            (((!$this->entity->isShowAll()) && isset($data[$key]) && (!in_array($data[$key],$element['#options']))) || ($this->entity->isShowAll()))
            ) {
              $tree[$key]['rows']['other'] = $other['childs'][$key.'_other']['rows'][$key.'_other'];
              $tree[$key]['rows']['other']['selected'] = (isset($data[$key]) ? !( in_array($data[$key],$element['#options']) ):false);
              $subpoints += $other['subtotal'];
            }
          $tree[$key]['footer'] = [[
            'data'=>t('@element: %subtotal Point(s)', [ '@element'=>$tree[$key]['title'], '%subtotal'=> $subpoints]),
            'attributes'=>' colspan=3'
          ]];
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
        case 'webform_likert':
          foreach($element['#questions'] as $qkey => $qvalue ) {
            $tree[$key]['rows'][$qkey] = [
              'question' => ['data'=>$qvalue, 'attributes'=>' colspan=3']
            ];
            foreach($element['#answers'] as $ankey => $anvalue ) {
              if (((!$this->entity->isShowAll()) && ( isset($data[$key]) && isset($data[$key][$qkey]) && ($ankey == $data[$key][$qkey]) )) ||
                    ($this->entity->isShowAll())) {
                $tree[$key]['rows'][$qkey.'-'.$ankey] = [
                  'selected' => (isset($data[$key][$qkey]) ? $ankey == $data[$key][$qkey] : false),
                  'answer' => $anvalue,
                  'points' => (isset($settings[$key][$qkey.'-'.$ankey]['points']) ? $settings[$key][$qkey.'-'.$ankey]['points'] : 0),
                  'isgood' => isset($data[$key]) && isset($data[$key][$qkey]) && ($ankey == $data[$key][$qkey]),
                ];
              }
              if ( isset($data[$key]) && isset($data[$key][$qkey]) && ($ankey == $data[$key][$qkey]) ) {
                $subpoints += $settings[$key][$qkey.'-'.$ankey]['points'];
              }
            }
          }
          $tree[$key]['footer'] = [[
            'data'=>t('@element: %subtotal Point(s)', [ '@element'=>$tree[$key]['title'], '%subtotal'=> $subpoints]),
            'attributes'=>' colspan=3'
          ]];
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
        case 'number':
        case 'range':
          $goodvalues = [];
          $isgood = false;
          if (isset($settings[$key.'_t'][$key]['answer'])) {
            $goodvalues = explode(',',$settings[$key.'_t'][$key]['answer']);
          }
          array_walk($goodvalues,'trim');
          foreach ( $goodvalues as $value ) {
            $pos = strpos( $value, '-');
            if ($pos == false) {
              if (isset($data[$key]) && ($value == $data[$key]) ) {
                $isgood = true;
              }
            }
            else {
              $range = explode('-',$value);
              array_walk($range,'trim');
              if ( isset($data[$key]) && ($range[0]<=$data[$key]) && ($data[$key] <= $range[1]) ) {
                $isgood = true;
              }
            }
          }
          $tree[$key]['rows'][$key] = [
            'selected' => $isgood,
            'answer' => (isset($data[$key]) ? $data[$key] : $element['#title']),
            'points' => ($isgood ? $settings[$key.'_t'][$key]['points'] : 0 ),
            'isgood' => $isgood,
          ];
          if ( $isgood ) {
              $subpoints += $settings[$key.'_t'][$key]['points'];
          }
          $tree[$key]['footer'] = [[
            'data'=>t('@element: %subtotal Point(s)', [ '@element'=>$tree[$key]['title'], '%subtotal'=> $subpoints]),
            'attributes'=>' colspan=3'
          ]];
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
        case 'webform_table_sort':
        case 'webform_tableselect_sort':
          $lastgood = 0;
          $tree[$key]['header'] = [ '',t('answer'),t('is at good place?') ];
          if (isset($data[$key])) {
            $i = 0;
            if (isset($settings[$key.'_order'])) {
              foreach ($settings[$key.'_order'] as $skey => $sval) {
                $setting[++$i] = $skey;
              }
            }
            $i = 0;
            foreach($data[$key] as $optkey => $optvalue ) {
              $tree[$key]['rows'][++$i] = [
                  'order' => $i,
                  'answer' => $element['#options'][$optvalue],
                  'points' => (isset($settings[$key][$key.'-'.$lastgood]) ? $settings[$key][$key.'-'.$lastgood]['points'] : 0),
                  'isgood' =>( isset($setting[$i]) ? ($setting[$i]==$optvalue ? t('true') : t('false'))  : ''),
              ];
              if (($setting[$i]==$optvalue) && ($lastgood==($i-1))){
                $lastgood = $i;
              }
            }
            if ( ($lastgood!=0) && isset($settings[$key][$key.'-'.$lastgood]['points']) ) {
              $subpoints += $settings[$key][$key.'-'.$lastgood]['points'];
            }
          }
          $tree[$key]['footer'] = [[
            'data' => t('The first %n item(s) ordered well: %points Point(s)', [
              '%n' => $lastgood,
              '%points' => $subpoints
            ]),
            'attributes' => ' colspan=3'
          ]];
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
          default:
            $tree[$key] = [];
            break;
        }
        $points += $subpoints;
      }
    }
    return [
      'childs' => $tree,
      'subtotal' => $points,
      'elements_points' => $elements_points,
    ];
  }

  /**
   *   Creates content form - the evaluation page
   *
   *     - @params comes from route: evalwf/[webform:id]/[webform_submission:sid]
   *     or from urlquery: evalwf?wfid=[webform:id]&sid=[webform_submission:sid]
   *
   *   @param Webform weform                         - the selected webform
   *   @param WebformSubmission webform_submission   - the selected webform_submission
   *   @param boolean isadmin                        - true if admin side called this
   *   @returns form array
   */
  public function content( Webform $webform = null, WebformSubmission $webform_submission = null, $isadmin = false ) {
    $evaluationismodified = false;
    $form = $this->LoadEntity($webform, $webform_submission);
    $query = \Drupal::entityQuery('evalwf_config')->accessCheck(TRUE)->condition('id', 'evalwf_config');
    $result = $query->execute();
    if ($result) $this->entityconfig = \Drupal::entityTypeManager()->getStorage('evalwf_config')->load('evalwf_config');

    if ($this->entity) {
      $form['title']=array(
        '#type' => 'markup',
        '#markup' => '<h1>'. $this->entity->getEvaluationTitle() .'</h1>',
      );
      $form[]=array(
        '#type' => 'markup',
        '#markup' => $this->entity->getEvaluationIntro(),
        '#suffix' => '<br><br>',
      );

      if (isset($this->sid) && isset($this->wfid) && !empty($this->entity->getElementsData())) {
        $result = $this->getElementsForm($this->webform_elements, $this->webform_submission->getData(), unserialize($this->entity->getElementsData()));
        $elements_form[] = $result['form'];
        $elements_form[] = array(
          '#type' => 'markup',
          '#markup' => '<h3><b><u>'.t('Total: ').$result['points'].' points</u></b></h3>',
        );
        $conditions_form = $this->getConditionsForm( array_merge( array('AllTotal' => $result['points']), $result['elements_points']) );
        $conditions = $this->getConditionsData( array_merge( array('AllTotal' => $result['points']), $result['elements_points']));
        if ($conditions) {
          $conditions[0]['title'] = t('Continue').' >>>';
          $conditions[0]['url'] = Url::fromUri($conditions[0]['url'])->setAbsolute(TRUE)->toString();
        }

        $elements = $this->buildelementstree(
          $this->webform_elements,
          $this->webform_submission->getData(),
          unserialize($this->entity->getElementsData())
        );
        $elements['childs']['AllTotal'] = [
          'title' => t('Your final result: %subtotal points', [ '%subtotal' => $elements['subtotal']]),
        ];
        $build = [
          '#theme' => 'evalwf_theme_hook',
          '#settings' => [
            'title' => $this->entity->getEvaluationTitle(),
            'html' => $this->entity->getEvaluationIntro(),
          ],
          '#elements' => $elements['childs'],
          '#conditions' => $conditions,
          '#attached' => [
            'library' => [
              'evalwf/evalwf-resultpage',
            ],
          ],
        ];

        $query = \Drupal::entityQuery('evalwf_evaluation')->accessCheck(FALSE)->condition('id', $this->sid);
        $res = $query->execute();
        if ($res) {
          $evaluation = \Drupal::entityTypeManager()->getStorage('evalwf_evaluation')->load($this->sid);
          $checksum = $evaluation->getCheckSum();
          if (md5(serialize($this->webform_elements))!=$checksum['webform']) {
            \Drupal::messenger()->addMessage( t('The webform elements has been modified, this result may different than the previously saved evaluation.'),'warning');
          }
          if (md5(serialize($this->webform_submission->getData()))!=$checksum['submission']) {
            \Drupal::messenger()->addMessage( t('The webform submission has been modified, this result may different than the previously saved evaluation.'),'warning');
          }
          if (md5($this->entity->getElementsData() . $this->entity->getConditionsData())!=$checksum['evalwf_setting']) {
            \Drupal::messenger()->addMessage( t('The evaluation settings has been changed, this result may different than the previously saved evaluation.'),'warning');
          }
        }
        else {
          $evaluation = new EvalWFEvaluation(
            [
              'id' => $this->sid,
              'wfid' => $this->wfid,
              'sid' => $this->sid,
              'total_points' => $result['points'],
              'elements_points' => serialize($result['elements_points']),
              'conditions_data' =>
                serialize($this->getConditionsData( array_merge( array('AllTotal' => $result['points']), $result['elements_points']) )),
              'sent' => false,
              'timestamp' => date('Y-m-d H:i:s', time()),
              'uid' => ( isset($this->entityconfig) && $this->entityconfig->isEvalAsUser() ?
                  $this->webform_submission->getOwner()->id() :
                  \Drupal::currentUser()->id()
                ),
              'checksum' => serialize(
                array(
                  'webform' => md5(serialize($this->webform_elements)),
                  'submission' => md5(serialize($this->webform_submission->getData())),
                  'evalwf_setting' => md5($this->entity->getElementsData() . $this->entity->getConditionsData()),
                  'evalwf_evaluation' => md5(serialize($result['elements_points']) . serialize($conditions_form)),
                )
              ),
            ],
            'evalwf_evaluation'
          );
          $evaluationismodified = true;
        }

        if ((($this->entity->isSendInEmail() || $this->entity->isSendToUser()) && !$evaluation->isSent()) &&
            ( ($isadmin && isset($this->entityconfig) && $this->entityconfig->isAdminSendMail()) || (!$isadmin) ) ){

          if ( $this->entity->isRenderAsForm(true) ) {
            if (!$this->entity->isIncludePTitle(true))
              unset($form['title']);
            $body = array_merge( $form, $elements_form );
            if (isset($conditions_form[0])) {
              $conditions_form[0]['#url']->setAbsolute(TRUE);
              $body = array_merge( $body, $conditions_form );
            }
          }
          else {
            if (!$this->entity->isIncludePTitle(true))
              unset( $build['#settings']['title'] );
            $body = $build;
          }

          $submitter = ( $isadmin ? $this->webform_submission->getOwnerId() : \Drupal::currentUser()->id() );
          $sendto = '';
          $sendto .= ($this->entity->isSendInEmail() ? $this->entity->getCollectorEmailAddress() : '' );
          $useraddr = '';
          $useraddr .= ($this->entity->isSendToUser() ? \Drupal\user\Entity\User::load( $submitter )->getEmail() : '');
          $sendto .= (($sendto=='' || $useraddr=='') ? '' :  ',') . $useraddr;

         if ( $sendto && $body ) {
            if ( isset($this->entityconfig) && $this->entityconfig->getSubject() ) {
              $subject = (new FormattableMarkup(
                  $this->entityconfig->getSubject(),
                  [ '@wfid' => $this->wfid, '@sid' => $this->sid, ]
                ))->__toString();
            }
            else {
              $subject = t( 'EvalWF - Evaluation of submission @wfid-@sid', [ '@wfid' => $this->wfid, '@sid' => $this->sid, ]);
            }
            $this->send_mail(
              ( isset($this->entityconfig) && $this->entityconfig->getSenderName() ?
                  $this->entityconfig->getSenderName() :
                  t('EvalWF at @site', [ '@site' => \Drupal::config('system.site')->get('name')])
              ),
              ( isset($this->entityconfig) && $this->entityconfig->getSender() ?
                  $this->entityconfig->getSender() :
                  \Drupal::config('system.site')->get('mail')
              ),
              $sendto,
              $subject,
              \Drupal::service('renderer')->render($body)->__toString(),
              ( isset($this->entityconfig) ?
                  $this->entityconfig->getMailer() :
                  'drupalmail'
              )
            );
          $evaluation->setSent( true );
          $evaluation->setSentTo( $sendto );
          $evaluationismodified = true;
          }
        }
        if ($evaluationismodified)
          $evaluation->save();
        if ($this->entity->isShowEvaluationPage()) {
          $form = array_merge( $form, $elements_form );
        }
        if ($conditions_form) $form = array_merge( $form, $conditions_form );
        if (!$this->entity->isIncludePTitle()) {
          unset($form['title']);
        }
      }
      else {
        if (isset($this->sid) && isset($this->wfid) && empty($this->entity->getElementsData())) {
          $form[]=array(
            '#type' => 'markup',
            '#markup' => t('Evaluation data currenly not available, please come back later...').'<br>',
          );
          if ($isadmin) {
            $button = array(
              '#type' => 'link',
              '#title' => t('Edit element settings'),
              '#url' => Url::fromRoute('entity.evalwf.settings_form.elements', ['webform' => $this->wfid,] ),
            );
            \Drupal::messenger()->addMessage(
              t('Evaluation settings of this webform haven\'t completed yet. Please fill and save Elements settings page...'),
              'warning'
            );
             \Drupal::messenger()->addMessage( $button, 'warning' );
          }
        }
        else {
          $form[]=array(
            '#type' => 'markup',
            '#markup' => t('NO DATA!').'<br>',
          );
        }
      }

      if ($this->entity->isDirectJump() && (!$isadmin)) {
        $conditions = unserialize($evaluation->getConditionsData());
        $responseurl = Url::fromUri($conditions[0]['url'])->setAbsolute(TRUE)->toString();
        if ( \Drupal::VERSION >= "9" ) {
          $response = new TrustedRedirectResponse( $responseurl, '302');
        }
        else {
          if (Url::fromUri($conditions[0]['url'])->setAbsolute(TRUE)->isExternal()) {
            return $conditions_form;
          }
          $response = new RedirectResponse( $responseurl, '302');
        }
        return $response;
      }
      else {
        if (!($this->entity->isRenderAsForm())) {
          if (!$this->entity->isIncludePTitle())
            unset( $build['#settings']['title'] );
          return $build;
        }
      }
    }
    else {
      $form[]=array(
        '#type' => 'markup',
        '#markup' => t('No evaluation settings were set to this request!').'<br>'
      );
      if ($isadmin) {
        $button = array(
          '#type' => 'link',
          '#title' => t('Add evaluation settings'),
          '#url' => Url::fromRoute('entity.evalwf.withpath.settings_form', ['webform' =>  $this->wfid,] ),
        );
        \Drupal::messenger()->addMessage(
          t('No Evaluation was set to this webform... Please add evaluation settings first...'),
          'warning'
        );
        \Drupal::messenger()->addMessage( $button, 'warning' );
      }
    }
    return $form;
  }

  /**
   *   Sends an email with these parameters
   *
   *   @param string from    - the sender email address
   *   @param string to      - the recipients
   *   @param string subject - the email subject
   *   @param string body    - the email body
   *   @param string mailer  - the mailmanager: phpmail/drupalmail
   */
  function send_mail($fromname, $from, $to, $subject, $body, $mailer = null ) {
    if ($this->entity->isRenderAsForm(true)) {
      $css = '<style>caption {text-align: left;font-weight: bold;font-size: 18px;}table {width:95%;padding-bottom:20px;}thead {background-color: #eee;}tfoot {background-color: #bbb;}.details-wrapper {border: 1px solid #aaa;margin-left: 20px;padding-left: 20px;}</style>';
    }
    else {
      $cssfile = $_SERVER['DOCUMENT_ROOT'] . base_path() . \Drupal::theme()->getActiveTheme()->getPath() . 'css/evalwf-resultpage-theme.css';
      if (file_exists($cssfile)) {
        $css = file_get_contents( $cssfile );
      }
      else {
        $cssfile = $_SERVER['DOCUMENT_ROOT'] . base_path() . \Drupal::service('extension.list.module')->getPath('evalwf') . '/css/evalwf-resultpage-theme.css';
        if (file_exists($cssfile)) {
          $css = file_get_contents( $cssfile );
        }
      }
      if ($css!==FALSE) {
        $css = '<style>'. $css .'</style>';
      }
    }
    switch ($mailer) {
      case 'phpmail':
        $mailManager = new PhpMail();
        $message['headers'] = array(
          'content-type' => 'text/html',
          'MIME-Version' => '1.0',
          'charset' =>  'UTF-8',
          'reply-to' => $from,
          'From' => $fromname." <".$from.">",
        );
        $message['to'] = $to;
        $message['subject'] = $subject;
        $message['body'] = '<html><head>'. $css . '</head><body>' . $body . '</body></html>';
        $mailManager->mail($message);
      break;
      case 'drupalmail':
      default:
        $mailManager = \Drupal::service('plugin.manager.mail');
        $langcode = \Drupal::currentUser()->getPreferredLangcode();
        $params['headers']['Content-Type'] = 'text/html; charset=UTF-8;';
        $params['sendername'] = $fromname;
        $params['sender'] = $from;
        $params['subject'] = $subject;
        $params['message'] = Html::escape('<html><head>'. $css . '</head><body>' . $body . '</body></html>');
        $mailManager->mail('evalwf', 'Evaluation', $to, $langcode, $params, null, $send = TRUE);
      break;
    }
  }

  /**
   *   Creates a table with evaluation data for one webform item
   *
   *   @param string title    - the title/caption of the table
   *   @param array  header   - the header items of the table
   *   @param array  rowsform - the rows in form array
   *   @param string total    - the total points of this item
   *   @returns form array
   */
  function createtable( $title, $header, $rowsform, $total ) {
    $tableform = array(
      '#type' => 'table',
      '#caption' => $title,
      '#header' => $header,
      '#prefix' => '<br>',
      '#footer' => array(
        'data' => array(
          array(
            'colspan' => 2,
            'data' => array(
              '#type' => 'markup',
              '#markup' => '<b>'. t('Subtotal:') .'</b>',
            ),
          ),
          array(
            'data' => array(
              '#type' => 'markup',
              '#markup' => '<b>'. $total .'</b>',
            ),
          ),
        )
      ),
    );
    if ( $rowsform ) { $tableform = array_merge( $tableform, $rowsform ); }
    return $tableform;
  }

  /**
   *   Creates a row with evaluation data for one webform item option/question/answer
   *
   *   @param array  item      - checkbox/radio item in form array
   *   @param string label     - the label for this option
   *   @param string value     - the points of this option
   *   @returns form array
   */
  function createtablerow( $item, $label, $value ) {
    $rowform = array(
      'item'   => $item,
      'answer' => array(
        '#type' => 'markup',
        '#markup' => $label,
      ),
      'points' => array(
        '#type' => 'markup',
        '#markup' => $value,
      ),
    );
    return $rowform;
  }

  /**
   *   Recursively creates a form with webform elements and evaluation datas
   *
   *   @param array  items      - webform elements
   *   @param array  data       - webform submission data
   *   @param array  settings   - the evaluation settings data
   *   @returns array
   *        array  form             -  elementsform data
   *        string points           -  total points of all elements
   *        array  elements_points  -  elements and its evaluated points
   */
  function getElementsForm( $items, $data, $settings ) {
    $points = 0;
    $elementsform = [];
    $elements_points = [];
    foreach ($items as $key => $item ) {
      if (isset($item['#type'])) {
      $optionsform = null;
      $subpoints = 0;
      switch ( $item['#type']) {
        case 'container':
        case 'fieldset':
        case 'details':
          $sub = $this->getElementsForm($item, $data, (isset($settings[$key]) ? $settings[$key] : null) );
          $elementsform[$key]=array(
            '#type' => 'details',
            '#title' => ( isset($item['#title']) ? $item['#title'] : '' ),
            '#tree' => TRUE,
            '#open' => TRUE,
            '#suffix' =>'<b>'. t('Subtotal of @key is: @subpoints points', [ '@key'=>$key, '@subpoints'=>$sub['points'] ]).'</b><br><br>',
          );
          $elementsform[$key][]= $sub['form'];
          $points += $sub['points'];
          $elements_points = array_merge( $elements_points, array( $key => $sub['points']) );
          $elements_points = array_merge( $elements_points, $sub['elements_points'] );
          break;
        case 'textarea':
        case 'textfield':
        case 'email':
        case 'webform_autocomplete':
          $elementsform[$key]=array(
            '#type' => 'markup',
            '#markup' => '<b>'. $item['#title'] .'</b>',
            '#suffix' => '<br>',
          );
          $elementsform[$key][$key]=
          [ array(
            '#type' => 'markup',
            '#markup' => ( isset($data[$key]) ? $data[$key] : null ),
            '#wrapper_attributes' => array(
              'colspan' => 2,
            ),
          )];
          break;
        case 'checkbox':
          $optionsform[$key]= $this->createtablerow(
            array(
              '#type' => 'checkbox',
              '#name' => $key,
              '#value' => ($data[$key]),
              '#checked' => ($data[$key]),
              '#attributes' => array('disabled' => TRUE),
            ),
            $item['#title'],
            $settings[$key.'_t'][$key]['points']
          );
          if ( isset($data[$key]) && ($data[$key]) && isset($settings[$key.'_t'][$key]['points']) ) {
              $subpoints += $settings[$key.'_t'][$key]['points'];
          }
          $elementsform[$key]= $this->createtable(
            $item['#title'],
            array( ['width' => 10, 'data' => ''],t('answer'),t('points') ),
            $optionsform,
            $subpoints
          );
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
        case 'webform_checkboxes_other':
          $otherval = (isset($data[$key]) ? $data[$key] : [] );
          $ov = null;
          foreach($item['#options'] as $optkey => $optvalue ) {
            foreach( $otherval as $okey => $oval ) {
              if ($oval == $optkey) $otherval[$okey] = null;
            }
          }
          foreach( $otherval as $okey => $oval ) {
            if ($oval!=null )$ov = $oval;
          }
          $other = $this->getElementsForm(
            [ $key.'_other' => array(
              '#type' => ( isset($item['#other__type']) ? $item['#other__type'] : 'textfield' ),
              '#title' => ( isset($item['#other__title']) ? $item['#other__title'] : null ),
              '#description' => ( isset($item['##other__description']) ? $item['##other__description'] : null ),
              '#size' => ( isset($item['#other__size']) ? $item['#other__size'] : null ),
            )],
            [$key.'_other' => $ov],
            ( isset($settings[$key][$key.'_other']) ? [$key.'_other_t' => [ $key.'_other' => $settings[$key][$key.'_other']]] : [] )
          );
        case 'checkboxes':
        case 'tableselect':
          foreach($item['#options'] as $optkey => $optvalue ) {
            if (((!$this->entity->isShowAll()) && isset($data[$key]) && ( in_array($optkey,$data[$key]) )) || ($this->entity->isShowAll())) {
              $optionsform[$optkey]= $this->createtablerow(
                array(
                  '#type' => 'checkbox',
                  '#name' => $key,
                  '#value' => (isset($data[$key]) ? in_array($optkey,$data[$key]) : FALSE ),
                  '#checked' => (isset($data[$key]) ? in_array($optkey,$data[$key]) : FALSE ),
                  '#attributes' => array('disabled' => TRUE),
                ),
                $optvalue,
                (isset($settings[$key][$optkey]['points']) ? $settings[$key][$optkey]['points'] : 0 )
              );
            }
            if ( isset($data[$key]) && in_array($optkey,$data[$key]) && isset($settings[$key][$optkey]['points'])  ) {
              $subpoints += $settings[$key][$optkey]['points'];
            }
          }
          if (
            (isset($other['form'][$key.'_other'][$key.'_other'])) &&
            (((!$this->entity->isShowAll()) && isset($data[$key]) && ( $ov!=null )) || ($this->entity->isShowAll()))
            ) {
            if (isset($other['form'][$key.'_other'][$key.'_other']['item'])) {
              $optionsform['other'] = $other['form'][$key.'_other'][$key.'_other'];
              $optionsform['other']['item']['#checked'] = ($ov!=null);
              $optionsform['other']['item']['#value'] = ($ov!=null);
            }
            else {
              $optionsform = array_merge( $optionsform, array (
                [
                  'item' => array(
                    '#type' => 'checkbox',
                    '#name' => $key.'_other',
                    '#value' => ($ov!=null),
                    '#checked' => ($ov!=null),
                    '#attributes' => array('disabled' => TRUE),
                  ),
                  $other['form'][$key.'_other'][$key.'_other'][0],
                ]
              ));
            }
            if (isset($other['points']))
              $subpoints += $other['points'];
          }
          $elementsform[$key]= $this->createtable(
            $item['#title'],
            array( ['width' => 10, 'data' => ''],t('answer'),t('points') ),
            $optionsform,
            $subpoints
          );
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
        case 'webform_radios_other':
        case 'webform_select_other':
          $other = $this->getElementsForm(
            [ $key.'_other' => array(
              '#type' => ( isset($item['#other__type']) ? $item['#other__type'] : 'textfield' ),
              '#title' => ( isset($item['#other__title']) ? $item['#other__title'] : null ),
              '#description' => ( isset($item['##other__description']) ? $item['##other__description'] : null ),
              '#size' => ( isset($item['#other__size']) ? $item['#other__size'] : null ),
            )],
            [$key.'_other' => (isset($data[$key]) && (!in_array($data[$key],$item['#options'])) ? $data[$key] : null )],
            ( isset($settings[$key][$key.'_other']) ? [$key.'_other_t' => [ $key.'_other' => $settings[$key][$key.'_other']]] : [] )
          );
        case 'radios':
          foreach($item['#options'] as $optkey => $optvalue ) {
            if (((!$this->entity->isShowAll()) && isset($data[$key]) && ( $optkey==$data[$key] )) || ($this->entity->isShowAll())) {
              $optionsform[$optkey]= $this->createtablerow(
                array(
                  '#type' => 'radio',
                  '#name' => $key,
                  '#return_value' => $optkey,
                  '#value' => ( (isset($data[$key]) && ($optkey==$data[$key])) ? $data[$key] : null),
                  '#attributes' => array('disabled' => TRUE),
                ),
                $optvalue,
                (isset($settings[$key][$optkey]['points']) ? $settings[$key][$optkey]['points'] : 0 )
              );
            }
            if ( isset($data[$key]) && ($optkey==$data[$key]) && isset($settings[$key][$optkey]['points'])) {
              $subpoints += $settings[$key][$optkey]['points'];
            };
          }
          if (
            (isset($other['form'][$key.'_other'][$key.'_other'])) &&
            (((!$this->entity->isShowAll()) && isset($data[$key]) && ( !in_array($data[$key],$item['#options']) )) || ($this->entity->isShowAll()))
            ) {
            if (isset($other['form'][$key.'_other'][$key.'_other']['item'])) {
              $optionsform['other'] = $other['form'][$key.'_other'][$key.'_other'];
              $optionsform['other']['item']['#checked'] = (isset($data[$key]) ? !( in_array($data[$key],$item['#options']) ):false);
              $optionsform['other']['item']['#value'] = (isset($data[$key]) ? !( in_array($data[$key],$item['#options']) ):false);
            }
            else {
              $optionsform = array_merge( $optionsform, array (
                [
                  'item' => array(
                    '#type' => 'checkbox',
                    '#name' => $key.'_other',
                    '#value' => (isset($data[$key]) ? !( in_array($data[$key],$item['#options']) ):false),
                    '#checked' => (isset($data[$key]) ? !( in_array($data[$key],$item['#options']) ):false),
                    '#attributes' => array('disabled' => TRUE),
                  ),
                  $other['form'][$key.'_other'][$key.'_other'][0],
                ]
              ));
            }
            if (isset($other['points']))
              $subpoints += $other['points'];
          }
          $elementsform[$key]= $this->createtable(
            $item['#title'],
            array( ['width' => 10, 'data' => ''],t('answer'),t('points') ),
            $optionsform,
            $subpoints
          );
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
        case 'select':
          foreach($item['#options'] as $optkey => $optvalue ) {
            if (((!$this->entity->isShowAll()) && isset($data[$key]) && ( $optkey==$data[$key] )) || ($this->entity->isShowAll())) {
              $optionsform[$optkey]= $this->createtablerow(
                array(
                  '#type' => 'checkbox',
                  '#name' => $key,
                  '#value' => $optkey == $data[$key],
                  '#checked' => $optkey == $data[$key],
                  '#attributes' => array('disabled' => TRUE),
                ),
                $optvalue,
                (isset($settings[$key][$optkey]['points']) ? $settings[$key][$optkey]['points'] : 0 )
              );
            }
            if ( isset($data[$key]) && ($optkey == $data[$key]) && isset($settings[$key][$optkey]['points']) ) {
              $subpoints += $settings[$key][$optkey]['points'];
            }
          }
          $elementsform[$key]= $this->createtable(
            $item['#title'],
            array( ['width' => 10, 'data' => ''],t('answer'),t('points') ),
            $optionsform,
            $subpoints
          );
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
        case 'webform_likert':
          foreach($item['#questions'] as $qkey => $qvalue ) {
            $optionsform[$qkey] = array(
              'item' => array(
                '#type' => 'markup',
                '#markup' => $qvalue,
                '#wrapper_attributes' => array(
                  'colspan' => 3,
                ),
              ),
            );
            foreach($item['#answers'] as $ankey => $anvalue ) {
              if (((!$this->entity->isShowAll()) && ( isset($data[$key]) && isset($data[$key][$qkey]) && ($ankey == $data[$key][$qkey]) )) ||
                    ($this->entity->isShowAll())) {
                $optionsform[$qkey.'-'.$ankey]= $this->createtablerow(
                  array(
                    '#type' => 'checkbox',
                    '#name' => $qkey,
                    '#value' => (isset($data[$key][$qkey]) ? $ankey == $data[$key][$qkey] : false),
                    '#checked' => (isset($data[$key][$qkey]) ? $ankey == $data[$key][$qkey] : false),
                    '#attributes' => array('disabled' => TRUE),
                  ),
                  $anvalue,
                  (isset($settings[$key][$qkey.'-'.$ankey]['points']) ? $settings[$key][$qkey.'-'.$ankey]['points'] : 0)
                );
              }
              if ( isset($data[$key]) && isset($data[$key][$qkey]) && ($ankey == $data[$key][$qkey]) && isset($settings[$key][$qkey.'-'.$ankey]['points']) ) {
                $subpoints += $settings[$key][$qkey.'-'.$ankey]['points'];
              }
            }
          }
          $elementsform[$key]= $this->createtable(
            $item['#title'],
            array( ['width' => 10, 'data' => ''],t('answer'),t('points') ),
            $optionsform,
            $subpoints
          );
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
        case 'number':
        case 'range':
          $goodvalues = [];
          $isgood = false;
          if (isset($settings[$key.'_t'][$key]['answer'])) {
            $goodvalues = explode(',',$settings[$key.'_t'][$key]['answer']);
          }
          array_walk($goodvalues,'trim');
          foreach ( $goodvalues as $value ) {
            $pos = strpos( $value, '-');
            if ($pos == false) {
              if (isset($data[$key]) && ($value == $data[$key]) ) {
                $isgood = true;
              }
            }
            else {
              $range = explode('-',$value);
              array_walk($range,'trim');
              if ( isset($data[$key]) && ($range[0]<=$data[$key]) && ($data[$key] <= $range[1]) ) {
                $isgood = true;
              }
            }
          }
          $optionsform[$key]= $this->createtablerow(
            array(
              '#type' => 'checkbox',
              '#name' => $key,
              '#value' => $isgood,
              '#checked' => $isgood,
              '#attributes' => array('disabled' => TRUE),
            ),
            (isset($data[$key]) ? $data[$key] : $item['#title']),
            ($isgood ? $settings[$key.'_t'][$key]['points'] : 0 )
          );
          if ( $isgood ) {
              $subpoints += $settings[$key.'_t'][$key]['points'];
          }
          $elementsform[$key]= $this->createtable(
            $item['#title'],
            array( ['width' => 10, 'data' => ''],t('answer'),t('points') ),
            $optionsform,
            $subpoints
          );
          $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          break;
        case 'webform_table_sort':
        case 'webform_tableselect_sort':
          if (isset($data[$key])) {
            $i = 0;
            if (isset($settings[$key.'_order'])) {
              foreach ($settings[$key.'_order'] as $skey => $sval) {
                $setting[++$i] = $skey;
              }
            }
            $i = 0;
            $lastgood = 0;
            foreach($data[$key] as $optkey => $optvalue ) {
              $optionsform[$optkey]= $this->createtablerow(
                array(
                  '#type' => 'markup',
                  '#markup' => ++$i .'.',
                ),
                $item['#options'][$optvalue],
                ( isset($setting[$i]) ? ($setting[$i]==$optvalue ? t('true') : t('false'))  : '')
              );
              if (($setting[$i]==$optvalue) && ($lastgood==($i-1))){
                $lastgood = $i;
              }
            }
            if ( ($lastgood!=0) && isset($settings[$key][$key.'-'.$lastgood]) && isset($settings[$key][$key.'-'.$lastgood]['points']) ) {
              $subpoints += $settings[$key][$key.'-'.$lastgood]['points'];
            }
            $elementsform[$key]= $this->createtable(
              $item['#title'],
              array( ['width' => 10, 'data' => t('order')],t('answer'),t('is at good place?') ),
              $optionsform,
              $subpoints
            );
            $elements_points = array_merge( $elements_points, array( $key => $subpoints ) );
          }
          break;
        default:
          break;
      }
      $points += $subpoints;
    }
    }

    return array( 'form' => $elementsform, 'points' => $points, 'elements_points' => $elements_points);
  }

  /**
   *   Creates a form with evaluation conditions data (the continue link)
   *
   *   @param array  points      - elements and its evaluated points
   *   @returns form array
   */
  function getConditionsForm( $points ) {
    $conditions = null;
    if ( !empty($this->entity->getConditionsData()) )
      $conditions = unserialize($this->entity->getConditionsData())['table'];
    $form = [];
    $result = false;
    if ($conditions!=null) {
      foreach($conditions as $index => $condition ) {
        if ( is_numeric($index) && ($condition['telem']!='') && ($condition['url']!='') ) {
          if ((!$result) && ( $points[ $condition['telem'] ] >= $condition['lo'] ) && ( $points[ $condition['telem'] ] <= $condition['hi'] ) ) {
            $form[] = array(
              '#type' => 'link',
              '#title' => t('Continue').' >>>',
              '#url' => Url::fromUri($condition['url']) ,
              '#prefix' => '<br>',
            );
            $result = true;
          }
        }
      }
      if ((!$result) && (isset($conditions['elsecondition'])) && $conditions['elsecondition']['iselsecondition']) {
        $form[] = array(
          '#type' => 'link',
          '#title' => t('Continue').' >>>',
          '#url' => Url::fromUri($conditions['elsecondition']['elseurl']),
          '#prefix' => '<br>',
        );
      }
      return $form;
    }
  }

  /**
   *   Creates conditions data (the continue link data)
   *
   *   @param array  points      - elements and its evaluated points
   *   @returns form array
   */
  function getConditionsData( $points ) {
    $conditions = null;
    if ( !empty($this->entity->getConditionsData()) )
      $conditions = unserialize($this->entity->getConditionsData())['table'];
    $data = [];
    $result = false;
    if ($conditions!=null) {
      foreach($conditions as $index => $condition ) {
        if ( is_numeric($index) && ($condition['telem']!='') && ($condition['url']!='') ) {
          if ((!$result) && ( $points[ $condition['telem'] ] >= $condition['lo'] ) && ( $points[ $condition['telem'] ] <= $condition['hi'] ) ) {
            $data[] = array(
              'telem' => $condition['telem'],
              'lo' => $condition['lo'],
              'points' => $points[ $condition['telem'] ],
              'hi' => $condition['hi'],
              'url' => $condition['url'],
            );
            $result = true;
          }
        }
      }
      if ((!$result) && (isset($conditions['elsecondition'])) && ($conditions['elsecondition']['iselsecondition'])) {
        $data[] = array(
          'telem' => null,
          'lo' => null,
          'points' => null,
          'hi' => null,
          'url' => $conditions['elsecondition']['elseurl'],
        );
      }
      return $data;
    }
  }
}
